@@ -1,11 +1,12 @@
 # Copyright (C) 2013-2015 The python-bitcoinlib developers
+# Copyright (C) 2018-2019 The python-bitcointx developers
 #
-# This file is part of python-bitcoinlib.
+# This file is part of python-bitcointx.
 #
 # It is subject to the license terms in the LICENSE file found in the top-level
 # directory of this distribution.
 #
-# No part of python-bitcoinlib, including this file, may be copied, modified,
+# No part of python-bitcointx, including this file, may be copied, modified,
 # propagated, or distributed except according to the terms contained in the
 # LICENSE file.
 
@@ -14,10 +15,100 @@
 import hashlib
 import unittest
 
-from bitcointx.core import b2x, x
+import bitcointx
+from bitcointx import ChainParams
+from bitcointx.util import dispatcher_mapped_list
+from bitcointx.core import b2x, x, Hash160
 from bitcointx.core.script import CScript, IsLowDERSignature
 from bitcointx.core.key import CPubKey
-from bitcointx.wallet import *
+from bitcointx.wallet import (
+    CCoinAddressError as CBitcoinAddressError,
+    CCoinAddress,
+    CBitcoinAddress,
+    CBase58BitcoinAddress,
+    CBech32BitcoinAddress,
+    P2PKHCoinAddress,
+    P2SHCoinAddress,
+    P2WPKHCoinAddress,
+    P2WSHCoinAddress,
+    P2PKHBitcoinAddress,
+    P2SHBitcoinAddress,
+    P2WPKHBitcoinAddress,
+    P2WSHBitcoinAddress,
+    CBitcoinKey,
+)
+
+
+def test_address_implementations(test, paramclasses=None,
+                                 extra_addr_testfunc=lambda *args: False):
+    pub = CPubKey(x('0378d430274f8c5ec1321338151e9f27f4c676a008bdf8638d07c0b6be9ab35c71'))
+    if paramclasses is None:
+        paramclasses = bitcointx.get_registered_chain_params()
+    for paramclass in paramclasses:
+        with ChainParams(paramclass):
+            def recursive_check(aclass):
+                test.assertTrue(issubclass(aclass, CCoinAddress))
+
+                if extra_addr_testfunc(aclass, pub):
+                    pass
+                else:
+                    a = None
+
+                    if getattr(aclass, 'from_pubkey', None):
+                        a = aclass.from_pubkey(pub)
+                    elif getattr(aclass, 'from_redeemScript', None):
+                        a = aclass.from_redeemScript(
+                            CScript(b'\xa9' + Hash160(pub) + b'\x87'))
+                    else:
+                        assert len(dispatcher_mapped_list(aclass)) > 0,\
+                            ("dispatcher mapped list for {} "
+                                "must not be empty".format(aclass))
+
+                    if a is not None:
+                        spk = a.to_scriptPubKey()
+                        test.assertEqual(a, aclass.from_scriptPubKey(spk))
+                        a2 = aclass.from_bytes(a)
+                        test.assertEqual(bytes(a), bytes(a2))
+                        test.assertEqual(str(a), str(a2))
+                        a3 = aclass(str(a))
+                        test.assertEqual(bytes(a), bytes(a3))
+                        test.assertEqual(str(a), str(a3))
+
+                for next_aclass in dispatcher_mapped_list(aclass):
+                    recursive_check(next_aclass)
+
+            recursive_check(CCoinAddress)
+
+
+class Test_CCoinAddress(unittest.TestCase):
+
+    def test_address_implementations(self):
+        test_address_implementations(self)
+
+    def test_get_output_size(self):
+        pub = CPubKey(x('0378d430274f8c5ec1321338151e9f27f4c676a008bdf8638d07c0b6be9ab35c71'))
+        a = P2PKHCoinAddress.from_pubkey(pub)
+        self.assertEqual(P2PKHCoinAddress.get_output_size(), 34)
+        self.assertEqual(a.get_output_size(), 34)
+        a = P2WPKHCoinAddress.from_pubkey(pub)
+        self.assertEqual(P2WPKHCoinAddress.get_output_size(), 31)
+        self.assertEqual(a.get_output_size(), 31)
+        a = P2SHCoinAddress.from_redeemScript(
+            CScript(b'\xa9' + Hash160(pub) + b'\x87'))
+        self.assertEqual(P2SHCoinAddress.get_output_size(), 32)
+        self.assertEqual(a.get_output_size(), 32)
+        a = P2WSHCoinAddress.from_redeemScript(
+            CScript(b'\xa9' + Hash160(pub) + b'\x87'))
+        self.assertEqual(P2WSHCoinAddress.get_output_size(), 43)
+        self.assertEqual(a.get_output_size(), 43)
+
+    def test_scriptpubkey_type(self):
+        for l1_cls in dispatcher_mapped_list(CCoinAddress):
+            for l2_cls in dispatcher_mapped_list(l1_cls):
+                for l3_cls in dispatcher_mapped_list(l2_cls):
+                    spk_type = l3_cls.get_scriptPubKey_type()
+                    matched_cls = CCoinAddress.match_scriptPubKey_type(spk_type)
+                    self.assertTrue(l3_cls is matched_cls)
 
 
 class Test_CBitcoinAddress(unittest.TestCase):
@@ -25,7 +116,10 @@ class Test_CBitcoinAddress(unittest.TestCase):
         """Create CBitcoinAddress's from strings"""
 
         def T(str_addr, expected_bytes, expected_version, expected_class):
-            addr = CBitcoinAddress(str_addr)
+            addr = CCoinAddress(str_addr)
+            addr2 = CBitcoinAddress(str_addr)
+            self.assertEqual(addr, addr2)
+            self.assertEqual(type(addr), type(addr2))
             self.assertEqual(addr.to_bytes(), expected_bytes)
             self.assertEqual(addr.__class__, expected_class)
             if isinstance(addr, CBase58BitcoinAddress):
@@ -119,7 +213,7 @@ class Test_CBitcoinAddress(unittest.TestCase):
         with self.assertRaises(CBitcoinAddressError):
             CBitcoinAddress.from_scriptPubKey(scriptPubKey)
 
-    def test_to_redeemScript(self):
+    def test_to_redeemScript_ok(self):
         def T(str_addr, expected_scriptPubKey_hexbytes):
             addr = CBitcoinAddress(str_addr)
 
@@ -127,15 +221,22 @@ class Test_CBitcoinAddress(unittest.TestCase):
             self.assertEqual(b2x(actual_scriptPubKey),
                              expected_scriptPubKey_hexbytes)
 
-        T('31h1vYVSYuKP6AhS86fbRdMw9XHieotbST',
-          'a914000000000000000000000000000000000000000087')
-
         T('1111111111111111111114oLvT2',
           '76a914000000000000000000000000000000000000000088ac')
 
         T('bc1qw508d6qejxtdg4y5r3zarvary0c5xw7kv8f3t4',
           '76a914751e76e8199196d454941c45d1b3a323f1433bd688ac')
 
+    def test_to_redeemScript_fail(self):
+        def T(str_addr):
+            addr = CBitcoinAddress(str_addr)
+
+            with self.assertRaises(NotImplementedError):
+                addr.to_redeemScript()
+
+        T('31h1vYVSYuKP6AhS86fbRdMw9XHieotbST')
+        T('bc1qc7slrfxkknqcq2jevvvkdgvrt8080852dfjewde450xdlk4ugp7szw5tk9')
+
     def test_to_scriptPubKey(self):
         """CBitcoinAddress.to_scriptPubKey() works"""
         def T(str_addr, expected_scriptPubKey_hexbytes):
@@ -166,12 +267,10 @@ class Test_P2PKHBitcoinAddress(unittest.TestCase):
     def test_from_non_canonical_scriptPubKey(self):
         def T(hex_scriptpubkey, expected_str_address):
             scriptPubKey = CScript(x(hex_scriptpubkey))
-            addr = P2PKHBitcoinAddress.from_scriptPubKey(scriptPubKey)
-            self.assertEqual(str(addr), expected_str_address)
-
-            # now test that CBitcoinAddressError is raised with accept_non_canonical_pushdata=False
+            # now test that CBitcoinAddressError is raised, non-canonical
+            # pushdata is not allowed
             with self.assertRaises(CBitcoinAddressError):
-                P2PKHBitcoinAddress.from_scriptPubKey(scriptPubKey, accept_non_canonical_pushdata=False)
+                P2PKHBitcoinAddress.from_scriptPubKey(scriptPubKey)
 
         T('76a94c14000000000000000000000000000000000000000088ac', '1111111111111111111114oLvT2')
         T('76a94d1400000000000000000000000000000000000000000088ac', '1111111111111111111114oLvT2'),
@@ -179,17 +278,15 @@ class Test_P2PKHBitcoinAddress(unittest.TestCase):
 
         # make sure invalid scripts raise CBitcoinAddressError
         with self.assertRaises(CBitcoinAddressError):
-            P2PKHBitcoinAddress.from_scriptPubKey(x('76a94c14'))
+            P2PKHBitcoinAddress.from_scriptPubKey(CScript(x('76a94c14')))
 
     def test_from_bare_checksig_scriptPubKey(self):
         def T(hex_scriptpubkey, expected_str_address):
             scriptPubKey = CScript(x(hex_scriptpubkey))
-            addr = P2PKHBitcoinAddress.from_scriptPubKey(scriptPubKey)
-            self.assertEqual(str(addr), expected_str_address)
-
-            # now test that CBitcoinAddressError is raised with accept_non_canonical_pushdata=False
+            # test that CBitcoinAddressError is raised, we do not support
+            # bare checksig
             with self.assertRaises(CBitcoinAddressError):
-                P2PKHBitcoinAddress.from_scriptPubKey(scriptPubKey, accept_bare_checksig=False)
+                P2PKHBitcoinAddress.from_scriptPubKey(scriptPubKey)
 
         # compressed
         T('21000000000000000000000000000000000000000000000000000000000000000000ac', '14p5cGy5DZmtNMQwTQiytBvxMVuTmFMSyU')
@@ -202,7 +299,7 @@ class Test_P2PKHBitcoinAddress(unittest.TestCase):
 
         # odd-lengths are *not* accepted
         with self.assertRaises(CBitcoinAddressError):
-            P2PKHBitcoinAddress.from_scriptPubKey(x('2200000000000000000000000000000000000000000000000000000000000000000000ac'))
+            P2PKHBitcoinAddress.from_scriptPubKey(CScript(x('2200000000000000000000000000000000000000000000000000000000000000000000ac')))
 
     def test_from_valid_pubkey(self):
         """Create P2PKHBitcoinAddress's from valid pubkeys"""
@@ -243,12 +340,12 @@ class Test_P2PKHBitcoinAddress(unittest.TestCase):
             P2PKHBitcoinAddress.from_pubkey(CPubKey(x('0378d430274f8c5ec1321338151e9f27f4c676a008bdf8638d07c0b6be9ab35c72')))
 
 
-class Test_CBitcoinSecret(unittest.TestCase):
+class Test_CBitcoinKey(unittest.TestCase):
     def test(self):
         def T(base58_privkey, expected_hex_pubkey, expected_is_compressed_value):
-            key = CBitcoinSecret(base58_privkey)
+            key = CBitcoinKey(base58_privkey)
             self.assertEqual(b2x(key.pub), expected_hex_pubkey)
-            self.assertEqual(key.is_compressed, expected_is_compressed_value)
+            self.assertEqual(key.is_compressed(), expected_is_compressed_value)
 
         T('5KJvsngHeMpm884wtkJNzQGaCErckhHJBGFsvd3VyK5qMZXj3hS',
           '0478d430274f8c5ec1321338151e9f27f4c676a008bdf8638d07c0b6be9ab35c71a1518063243acd4dfe96b66e3f2ec8013c8e072cd09b3834a19f81f659cc3455',
@@ -258,7 +355,7 @@ class Test_CBitcoinSecret(unittest.TestCase):
           True)
 
     def test_sign(self):
-        key = CBitcoinSecret('5KJvsngHeMpm884wtkJNzQGaCErckhHJBGFsvd3VyK5qMZXj3hS')
+        key = CBitcoinKey('5KJvsngHeMpm884wtkJNzQGaCErckhHJBGFsvd3VyK5qMZXj3hS')
         hash = b'\x00' * 32
         sig = key.sign(hash)
 
@@ -275,7 +372,7 @@ class Test_CBitcoinSecret(unittest.TestCase):
         self.assertFalse(key.pub.verify(hash, sig[0:-4] + b'\x00\x00\x00\x00'))
 
     def test_sign_invalid_hash(self):
-        key = CBitcoinSecret('5KJvsngHeMpm884wtkJNzQGaCErckhHJBGFsvd3VyK5qMZXj3hS')
+        key = CBitcoinKey('5KJvsngHeMpm884wtkJNzQGaCErckhHJBGFsvd3VyK5qMZXj3hS')
         with self.assertRaises(TypeError):
             key.sign('0' * 32)
 
@@ -285,30 +382,30 @@ class Test_CBitcoinSecret(unittest.TestCase):
 
     def test_from_to_compressed(self):
         def T(keydata, compressed, uncompressed):
-            k = CBitcoinSecret.from_secret_bytes(x(keydata))
-            k_u = CBitcoinSecret.from_secret_bytes(x(keydata), False)
+            k = CBitcoinKey.from_secret_bytes(x(keydata))
+            k_u = CBitcoinKey.from_secret_bytes(x(keydata), False)
 
-            self.assertTrue(k.is_compressed)
+            self.assertTrue(k.is_compressed())
             self.assertEqual(k.pub, x(compressed))
 
-            k2 = CBitcoinSecret(str(k))
-            self.assertTrue(k2.is_compressed)
+            k2 = CBitcoinKey(str(k))
+            self.assertTrue(k2.is_compressed())
             self.assertEqual(k, k2)
 
             k = k.to_uncompressed()
             self.assertEqual(k, k_u)
             self.assertEqual(len(k), 32)
-            self.assertFalse(k.is_compressed)
+            self.assertFalse(k.is_compressed())
             self.assertEqual(k.pub, x(uncompressed))
 
-            k2 = CBitcoinSecret(str(k))
-            self.assertFalse(k2.is_compressed)
+            k2 = CBitcoinKey(str(k))
+            self.assertFalse(k2.is_compressed())
             self.assertEqual(k, k2)
 
             k = k.to_compressed()
             self.assertEqual(len(k), 33)
             self.assertEqual(k[-1], 1)
-            self.assertTrue(k.is_compressed)
+            self.assertTrue(k.is_compressed())
             self.assertEqual(k.pub, x(compressed))
 
         T('0de5306487851213f0aae1454f4e4449949a755802b60f6eb47906149395d080',
@@ -331,7 +428,7 @@ class Test_RFC6979(unittest.TestCase):
             (0xe91671c46231f833a6406ccbea0e3e392c76c167bac1cb013f6f1013980455c2, "There is a computer disease that anybody who works with computers knows about. It's a very serious disease and it interferes completely with the work. The trouble with computers is that you 'play' with them!", 0x1F4B84C23A86A221D233F2521BE018D9318639D5B8BBD6374A8A59232D16AD3D, "b552edd27580141f3b2a5463048cb7cd3e047b97c9f98076c32dbdf85a68718b279fa72dd19bfae05577e06c7c0c1900c371fcd5893f7e1d56a37d30174671f6")
         ]
         for vector in test_vectors:
-            secret = CBitcoinSecret.from_secret_bytes(x('{:064x}'.format(vector[0])))
+            secret = CBitcoinKey.from_secret_bytes(x('{:064x}'.format(vector[0])))
             encoded_sig = secret.sign(hashlib.sha256(vector[1].encode('utf8')).digest())
 
             assert(encoded_sig[0] == 0x30)
