@@ -7,13 +7,14 @@
 # It is subject to the license terms in the LICENSE file found in the top-level
 # directory of this distribution.
 #
-# No part of python-bitcoinlib, including this file, may be copied, modified,
+# No part of python-bitcointx, including this file, may be copied, modified,
 # propagated, or distributed except according to the terms contained in the
 # LICENSE file.
 
+import os
 import sys
-import ssl
-from bitcointx.core import b2x, BIP32_HARDENED_KEY_LIMIT
+from bitcointx.core import b2x
+from bitcointx.core.key import BIP32Path
 from bitcointx.base58 import Base58Error, UnexpectedBase58PrefixError
 from bitcointx.wallet import CBitcoinExtKey, CBitcoinExtPubKey
 
@@ -24,7 +25,7 @@ if __name__ == '__main__':
         sys.exit(-1)
 
     if len(sys.argv) == 2:
-        xkey = CBitcoinExtKey.from_seed(ssl.RAND_bytes(32))
+        xkey = CBitcoinExtKey.from_seed(os.urandom(32))
         print("generated xpriv: ", xkey)
     else:
         for cls in (CBitcoinExtKey, CBitcoinExtPubKey):
@@ -42,45 +43,44 @@ if __name__ == '__main__':
             print("ERROR: specified key does not appear to be valid")
             sys.exit(-1)
 
-    path = sys.argv[1]
-    if path.startswith('m'):
-        path = path[1:]
+    path_str = sys.argv[1]
 
-    numeric_path = []
-    for elt in path.split('/'):
-        if elt == '':
-            continue
+    if not path_str.startswith('m'):
+        # Allow users to omit path prefix
+        if path_str == '':
+            path_str = 'm'
+        else:
+            path_str = 'm/' + path_str.lstrip('/')
 
-        c = elt
-        hardened = 0
-        if c.endswith("'") or c.endswith('h'):
-            hardened = BIP32_HARDENED_KEY_LIMIT
-            c = c[:-1]
-        try:
-            n = int(c) + hardened
-        except ValueError:
-            print("ERROR: invalid element in the path:", elt)
-            sys.exit(-1)
+    path = BIP32Path(path_str)
 
+    if len(path) == 0:
+        # NOTE: xkey.derive_path() method will raise ValueError
+        # on empty path, to guard against bugs:
+        #   if there is nothing to derive, why call derive_path() ?
+        print('ERROR: nothing to derive, path is empty.')
+        sys.exit(-1)
+
+    for n in path:
         print("child number: 0x{:08x}".format(n))
         xkey = xkey.derive(n)
         if isinstance(xkey, CBitcoinExtKey):
             print("xpriv:", xkey)
 
             # Note:
-            # if xkey is CBitcoinExtKey, xkey.priv is CBitcoinSecret
-            #     CBitcoinSecret is in WIF format, and compressed
+            # if xkey is CBitcoinExtKey, xkey.priv is CBitcoinKey
+            #     CBitcoinKey is in WIF format, and compressed
             #     len(bytes(xkey.privkey)) == 33
             # if xkey is CExtKey, xkey.priv is CKey
             #     CKey is always 32 bytes
             #
-            # Standalone CBitcoinSecret key can be uncompressed,
+            # Standalone CBitcoinKey key can be uncompressed,
             # and be of 32 bytes length, but this is not the case
             # with xpriv encapsulated in CBitcoinExtKey - it is
             # always compressed there.
             #
             # you can always use xkey.priv.secret_bytes
-            # to get raw 32-byte secret data from both CBitcoinSecret and CKey
+            # to get raw 32-byte secret data from both CBitcoinKey and CKey
             #
             print("priv WIF:", xkey.priv)
             print("raw priv:", b2x(xkey.priv.secret_bytes))
